// #define	DEBUG

/*
RMEE01
A/D変換にMC3424を使ったE24系列基準の抵抗計（温度計モード付き）
(C) 2021 JH4VAj
*/

#include <math.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

#define	TITLE		"RMEE01"
#define	VERSION	"1.02"

#define	ADC_RES_BITS	18			/* A/D変換器の分解能（ビット: 12, 14, 16, 18）	*/




#define	SYS_MODE_UNDEF	-1
#define	SYS_MODE_NORMAL	0
#define	SYS_MODE_THERMO	1

#define	RREF_HIGH	10.0e3			/* 基準抵抗の値（High側） */
#define	RREF_LOW	100.0				/* 基準抵抗の値（Low側） */

#define	RMIN_HIGH	5.0					/* 測定下限抵抗値（High側） */
#define	RMAX_HIGH	2.0e6				/* 測定上限抵抗値（High側） */		

#define	RMIN_LOW	5.0e-2			/* 測定下限抵抗値（Low側） */
#define	RMAX_LOW	2.0e4				/* 測定上限抵抗値（Low側） */	

#define	DIFF_MIN	512					/* 測定差分下限（これ未満は試料が接続されていないと見なす（8倍値））	*/

/* サーミスタ: NXFT15XH103FA2B050												*/
/* https://akizukidenshi.com/catalog/g/gP-05251/			*/
#define TH_B	3380.0	/* サーミスタのB定数								*/
#define TH_T0	25.0	 	/* B定数の基準温度（セ氏）						*/	
#define TH_R0	10.0e3	/* 基準温度におけるサーミスタの抵抗値	*/
#define	KC		273.15 	/* 絶対温度セ氏換算値								*/

LiquidCrystal_I2C lcd(0x27, 16, 2);

const float e24[] =	 {1.0, 1.1, 1.2, 1.3, 1.5, 1.6, 1.8,
											2.0, 2.2, 2.4, 2.7,
											3.0, 3.3, 3.6, 3.9,
											4.3, 4.7, 5.1, 5.6,
											6.2, 6.8, 7.5, 8.2, 9.1
											};


byte  adc_config_reg_ch_sel;	// コンフィグレジスタ: C1, C0	
int		adc_read_wait_time;			// A/D変換器読出し待ち時間、ms	

float	vadc_lsb;								// 最小電圧（1ビット分の電圧、V）

long	val_ref;								// 基準抵抗値（A/D変換器読出し値）

long	val_adj_low;						// 測定補正値（コネクタやテストリードによる抵抗分、Low側限定）
float	r_adj;									// 測定補正値（抵抗値）

char	str_min_val_low[8], str_max_val_low[8], str_min_val_high[8], str_max_val_high[8];
char	*p_str_min_val, *p_str_max_val;

int	sys_mode;

float	temp_lowest, temp_highest;
boolean	reset_temp_lo_hi;

byte	led_heart_beat;


void heart_beat() {
	led_heart_beat = ~led_heart_beat;
	digitalWrite(LED_BUILTIN, led_heart_beat);
}


/*
char* read_vref(char *buf) {
	return(dtostrf((((float)analogRead(A3) / 1024.0) * 5.0), 5, 3, buf));
}
*/


// the setup function runs once when you press reset or power the board
void setup() {
	pinMode(LED_BUILTIN, OUTPUT);

	sys_mode = SYS_MODE_UNDEF;

	led_heart_beat = 0;
	
	heart_beat();

	print_version();
	
	setup_adc(ADC_RES_BITS);
	
	check_voltage();
	//delay(1500);		// 補正値読み出しで時間がかかるのでここはナシで

	heart_beat();

	set_str_min_max_val();

	val_adj_low = 0;
	r_adj = 0;
	
	sys_mode = check_sys_mode();	// 動作モードチェック、および、抵抗補正値設定

	temp_lowest = 1000.0;		// 温度の初期値はありえない値に設定
	temp_highest = -1000.0;	// これで最初は必ず計測値がLow, Highに設定される
	reset_temp_lo_hi = true;
}


void print_version() {
  Wire.begin();	// lcd.init()の中で多分やっていると思うが

  lcd.init();
  lcd.backlight();

	lcd.clear();
	lcd.print(TITLE);
	lcd.print(" Ver.");
	lcd.print(VERSION);
	lcd.setCursor(0, 1);
	lcd.print("(C}2021 JH4VAJ");
	delay(1000);
}


void set_str_min_max_val() {
	dtostrf(RMIN_LOW, 5, 2, str_min_val_low);	// 測定範囲下限を文字列化（表示用）

	if (RMAX_LOW < 1.0e3) {										// 測定版上限を文字列可（k、Mの接頭辞対応）
		dtostrf(RMAX_LOW, 6, 2, str_max_val_low);
	}
	else if (RMAX_LOW < 1.0e6) {
		dtostrf((RMAX_LOW / 1.0e3), 6, 2, str_max_val_low);
		strcat(str_max_val_low, "k");
	}
	else {
		dtostrf((RMAX_LOW / 1.0e6), 6, 2, str_max_val_low);
		strcat(str_max_val_low, "M");
	}
	
	dtostrf(RMIN_HIGH, 5, 2, str_min_val_high);	// 測定範囲下限を文字列化（表示用）

	if (RMAX_HIGH < 1.0e3) {										// 測定版上限を文字列可（k、Mの接頭辞対応）
		dtostrf(RMAX_HIGH, 6, 2, str_max_val_high);
	}
	else if (RMAX_HIGH < 1.0e6) {
		dtostrf((RMAX_HIGH / 1.0e3), 6, 2, str_max_val_high);
		strcat(str_max_val_high, "k");
	}
	else {
		dtostrf((RMAX_HIGH / 1.0e6), 6, 2, str_max_val_high);
		strcat(str_max_val_high, "M");
	}
}



void setup_adc(int adc_res_bits) {
	switch (adc_res_bits) {
    case 12:
      adc_config_reg_ch_sel = 0x00;
      adc_read_wait_time = 6;
      break;
    case 14:
      adc_config_reg_ch_sel = 0x04;
      adc_read_wait_time = 20;
      break;
    case 16:
      adc_config_reg_ch_sel = 0x08;
      adc_read_wait_time = 80;
      break;
    case 18:
    default:
      adc_config_reg_ch_sel = 0x0C;
      adc_read_wait_time = 300;
      break;
  }

	//vadc_lsb = 2.048 / (float)pow(2.0, (float)(adc_res_bits - 1));
	vadc_lsb = 2.048 / (float)((long)1 << (adc_res_bits - 1));
	
	adc_read_wait_time = 300;	/* 読出しが速すぎるとLCDの表示がチラつくので300msで統一	*/
}


void check_voltage() {
	float v18_adc;				// MCP3424の1.8V（実測値）
	char	str_v18_adc[8];
	float	vcc_mcu;				// ATmega328PのVcc
	char	str_vcc_mcu[8];

	v18_adc = (float)read_adc(1, 1) * vadc_lsb;
	dtostrf(v18_adc, 5, 3, str_v18_adc);
	
	vcc_mcu = (v18_adc / (float)analogRead(A3)) * 1024.0;	// v18_adcを元に逆算
	dtostrf(vcc_mcu, 5, 3, str_vcc_mcu);

	lcd.clear();
	lcd.print("Vcc:");
	lcd.print(str_vcc_mcu);
	lcd.print("V");
	lcd.setCursor(0, 1);
	lcd.print("V18:");
	lcd.print(str_v18_adc);
	lcd.print("V");
}



/*
 * Read From A/D Converter MC3424
 * ADRS: 0x68
 * 
 * ch: 1, 2, 3, 4
 * pga: 1, 2, 4, 8		// 内蔵アンプの倍率
 */
 
long read_adc(byte ch, byte pga) {
	const byte	adrs = 0x68;
	
  long  dat;
  byte  adc_config_reg, num_dat;
  int		i;

	num_dat = (ADC_RES_BITS == 18) ? 4 : 3;	// データ読出しの回数（18ビットの場合だけ4回。それ以外は3回）

  switch (ch) {
  	case 4:
  		adc_config_reg = adc_config_reg_ch_sel | 0x70;
  		break;
  	case 3:
  		adc_config_reg = adc_config_reg_ch_sel | 0x50;
  		break;
  	case 2:
  		adc_config_reg = adc_config_reg_ch_sel | 0x30;
  		break;
 		case 1:
 		default:
  		adc_config_reg = adc_config_reg_ch_sel | 0x10;
  		break;
  }

  switch (pga) {
  	case 8:
  		adc_config_reg |= 0x03;	// x8
  		break;
  	case 4:
  		adc_config_reg |= 0x02;	// x4
  		break;
  	case 2:
  		adc_config_reg |= 0x01;	// x2
  		break;
  	case 1:
  	default:
  		adc_config_reg |= 0x00;	// x1
  		break;
  }

  Wire.beginTransmission(adrs);
  Wire.write(adc_config_reg);
  Wire.endTransmission(false);

  delay(adc_read_wait_time);

  Wire.requestFrom(adrs, num_dat);

	dat = 0;
	for (i = 0; i < (num_dat - 1); i++) {
		dat <<= 8;
		dat |= Wire.read();
	}
	
	adc_config_reg = Wire.read();

	if ((adc_config_reg & 0x80) != 0x00) {
		print_error("ADC read error", "Please reset");
		
		while(1) {}	// エラー時は動作停止
	}

	if (dat < 0) {
		dat = 0;
	}

	if ((dat & ((long)1 << (ADC_RES_BITS - 1))) != 0) {	// if (MSB == HIGH) {dat = 0;}
  	dat = 0;																		// マイナスの値なら0を返す
	}

  return(dat);
}


/*
 * A/Dコンバータからの読出し（8倍値）
 *   ADCの値が小さい値なら8倍精度で読み出す
 */

long read_adc_8(byte ch) {
	long	dat;
	const long	thld_8 = 0x00000800;	// 8倍読出しを行う閾値
	
	dat = read_adc(ch, 1);		// A/Dコンバータから読み出す（1倍）

	if (dat > thld_8) {				// if (読み出した値が大きい)
		dat <<= 3;								// 8倍にする
	}
	else {
		dat = read_adc(ch, 8);		// アンプを有効化して8倍精度で読み出す
	}
	
	return(dat);
}


int check_sys_mode()
{
	long	val_ch2, val_ch3, v;
	int		i, ret;
	char  str_buf[8];

	ret = SYS_MODE_UNDEF;

	val_ref = read_adc_8(1);		// 一番最初のval_ref読出し（val_refの初期化）
	val_ref += read_adc_8(1);		// 二回読み出して平均
	val_ref >>= 1;

	heart_beat();
	
	val_ch2 = read_adc_8(2);
	val_ch2 = (val_ref < val_ch2) ? val_ref : val_ch2;
	
	/*
	val_ch3 = 0x7fffffff;				// Low chは4回読んで最小値を採用する（測定補正値用）
	for (i = 0; i < 4; i++) {
		v = read_adc_8(3);
		val_ch3 = (v < val_ch3) ? v : val_ch3;
		heart_beat();
	}
	*/

	v = 0;											// Low chは4回読んで平均値を採用する（測定補正値用）
	for (i = 0; i < 4; i++) {
		v += read_adc_8(3);
		heart_beat();
	}
	val_ch3 = v >> 2;

	if ((val_ref - val_ch2) < DIFF_MIN) {			// if (High側 == OPEN) （差がほとんどなければオープン）
		if ((val_ref - val_ch3) > DIFF_MIN) {		// 		if (Low側 != OPEN))
			val_adj_low = val_ch3;
			
			r_adj = ((float)val_adj_low / (float)(val_ref - val_adj_low)) * RREF_LOW;

			dtostrf(r_adj, 5, 3, str_buf);

			lcd.clear();
			lcd.print("adj: ");
			lcd.print(str_buf);
			lcd.write(0xf4);

			delay(1000);
		}
		
		ret = SYS_MODE_NORMAL;
	}
	else {
		ret = SYS_MODE_THERMO;
	}
	
	return(ret);
}


void read_val_ref() {		// val_ref読出し（8回分の平均）
	long v;

	v = val_ref << 3;
	v = v - val_ref;
	val_ref = (v + read_adc_8(1)) >> 3;
	
	return;
}


void do_normal() {
	long	val_tgt, val_diff_high, val_diff_low;
	float	r, r_max, r_min;
	char 	msg[18];

	/* Read reference (CH1) */
  read_val_ref();

	/* Read target (CH2: High)*/
	val_tgt = read_adc_8(2);
	val_diff_high = val_ref - val_tgt;
	val_diff_high = (val_diff_high < DIFF_MIN) ? 0 : val_diff_high;

	if (0 < val_diff_high) {
		r = (((float)val_tgt / (float)val_diff_high) * RREF_HIGH) - r_adj;
		r = (r < 0) ? 0 : r;
	}
	else {				// val_ref < val_tgt（A/Dコンバータの誤差など）
		r = RMAX_HIGH + 1.0;
	}

#ifdef DEBUG2
	lcd.clear();
	lcd.print(val_ref);
	lcd.setCursor(0, 1);
	lcd.print(val_tgt);
	lcd.setCursor(12, 1);
	lcd.print(val_diff_high);
	delay(1500);
#endif

	r_max = RMAX_HIGH;
	r_min = RMIN_HIGH;
	p_str_max_val = str_max_val_high;
	p_str_min_val = str_min_val_high;

	val_diff_low = 0;

  if (RMAX_HIGH < r) {	/* if (CH2 is open)	*/
  	/* Read target (CH3: Low) */
		val_tgt = read_adc_8(3);
		val_diff_low = val_ref - val_tgt;
		val_diff_low = (val_diff_low < DIFF_MIN) ? 0 : val_diff_low;
	
		if (val_diff_low >= DIFF_MIN) {		// if (Low側に何かつながっている）
			if (0 < val_diff_low) {
				r = (((float)val_tgt / (float)val_diff_low) * RREF_LOW) - r_adj;
				r = (r < 0) ? 0 : r;
			}
			else {				// val_ref < val_tgt（A/Dコンバータの誤差など）
				r = RMAX_LOW + 1.0;
			}

			r_max = RMAX_LOW;
			r_min = RMIN_LOW;
			p_str_max_val = str_max_val_low;
			p_str_min_val = str_min_val_low;
		}
  }

#ifdef DEBUG
	lcd.clear();
	lcd.print(val_diff_low);
	lcd.setCursor(0, 1);
	lcd.print(val_diff_high);
	delay(1000);
#endif

  if (r < r_min) {
		sprintf(msg, "Rx < %s", p_str_min_val);
		print_error(msg, "");
	}
	else if (r_max < r) {
		if ((val_diff_low >= DIFF_MIN) || (val_diff_high >= DIFF_MIN)) {
			sprintf(msg, "%s < Rx", p_str_max_val);
			print_error(msg, "");
		}
		else {
			sprintf(msg, "Open (too high)");
			print_error(msg, "");
		}
	}
	else {
		print_result(r);
	}

	return;
}




void print_error(char* msg1, char* msg2) {
	lcd.clear();
	lcd.print(msg1);
	lcd.setCursor(0, 1);
	lcd.print(msg2);

	return;
}



void print_result(float r) {
	const	float	r_thsld = 0.955;

	char  str_buf[10], *str_kM;
  float	kM, match_e24;
	float	multi, diff_0, diff_1, diff_kM, diff_cent;
  int		n_digi;
	int		i;
  
  if (r > (r_thsld * 1.0e6)) {
  	str_kM = "M";
  	kM = 1.0e6;
  }
  else if (r > (r_thsld * 1.0e3)) {
  	str_kM = "k";
  	kM = 1.0e3;
  }
  else {
  	str_kM = "";
  	kM = 1.0;
  }
  	
	if (r < (r_thsld * 1.0)) {
    multi = 0.1;
    n_digi = 1;
  }
	else if (r < (r_thsld * 10.0)) {
    multi = 1.0;
    n_digi = 1;
  }
	else if (r < (r_thsld * 100.0)) {
    multi = 10.0;
    n_digi = 2;
  }
	else if (r < (r_thsld * 1.0e3)) {
    multi = 100.0;
    n_digi = 3;
  }
	else if (r < (r_thsld * 10.0e3)) {
    multi = 1.0e3;
    n_digi = 1;
  }
	else if (r < (r_thsld * 100.0e3)) {
    multi = 10.0e3;
    n_digi = 2;
  }
	else if (r < (r_thsld * 1.0e6)) {
    multi = 100.0e3;
    n_digi = 3;
  }
	else if (r < (r_thsld * 10.0e6)) {
    multi = 1.0e6;
    n_digi = 1;
  }
	else if (r < (r_thsld * 100.0e6)) {
    multi = 10.0e6;
    n_digi = 2;
  }
	else {
  	multi = 100.0e6;
  	n_digi = 3;
  }

  diff_1 = (r - (e24[0] * multi));
  diff_0 = diff_1;

  for (i = 1; i < (sizeof(e24) / sizeof(float)); i++) {
    diff_1 = r - (e24[i] * multi);
    if (abs(diff_0) < abs(diff_1)) {
      break;
    }
    else {
      diff_0 = diff_1;
    }
  }

	match_e24 = (e24[i - 1] * multi) / kM;
	
	lcd.clear();

	const int digi_mes = 5;
	dtostrf((r / kM), digi_mes, (digi_mes - 1 - n_digi), str_buf);
  lcd.print(str_buf);
  lcd.print(str_kM);
  lcd.write(0xf4);

	diff_kM = diff_0 / kM;
  lcd.setCursor(9, 0);
  dtostrf(diff_kM, 7, 3, str_buf);
  lcd.print(str_buf);

	if (n_digi == 1) {
		dtostrf(match_e24, 3, 1, str_buf);
	}
	else {
		dtostrf(match_e24, n_digi, 0, str_buf);
	}
	
  lcd.setCursor(0, 1);
  lcd.print(str_buf);
  lcd.print(str_kM);
	lcd.print(" E24");

	diff_cent = (diff_kM / match_e24) * 100.0;
	dtostrf(diff_cent, 6, 2, str_buf);
  lcd.setCursor(9, 1);
  lcd.print(str_buf);
  lcd.print("%");

  return;
}



void do_thermo() {
	long	val_tgt, val_diff;
	float	r, r_ref;
	float temp;

	if (read_adc(3, 1) < ((long)1 << (ADC_RES_BITS - 2))) {	// Low側端子がショートされているか？
		reset_temp_lo_hi = true;
	}
	else {
		reset_temp_lo_hi = false;
	}
	
	/* Read reference (CH1) */
  read_val_ref();

	/* Read target (CH2)*/
	val_tgt = read_adc_8(2);
	val_diff = val_ref - val_tgt;

	r_ref = RREF_HIGH;

 	r = ((float)val_tgt / (float)val_diff) * r_ref;

 	temp = (1.0 / ((log(r / TH_R0) / TH_B) + 1.0 / (TH_T0 + KC))) - KC;

	if (reset_temp_lo_hi == true) {
		temp_lowest = temp;
		temp_highest = temp;
	}
	else {
 		if (temp < temp_lowest) {
 			temp_lowest = temp;
 		}
	
 		if (temp_highest < temp) {
 			temp_highest = temp;
 		}
	}

 	lcd.clear();

	lcd.print("Now: ");
 	lcd.print(temp);
  lcd.write(0xdf);
  lcd.write(0x43);

  lcd.setCursor(0, 1);
  lcd.print("L:");
  if (reset_temp_lo_hi == true) {
  	lcd.print("---");
  }
  else{
 		lcd.print(temp_lowest);
  }
 	
 	lcd.setCursor(8, 1);
  lcd.print("H:");
  if (reset_temp_lo_hi == true) {
  	lcd.print("---");
  }
  else{
 		lcd.print(temp_highest);
  }

  return;
}



// the loop function runs over and over again forever
void loop() {
	heart_beat();

	if (sys_mode == SYS_MODE_NORMAL) {
		do_normal();
	}
	else {
		do_thermo();
	}
}
